package org.netkernel.lang.dpml.endpoint;

import java.io.IOException;

import org.netkernel.lang.dpml.ast.IArgument;
import org.netkernel.lang.dpml.ast.IIdentifierVariable;
import org.netkernel.lang.dpml.ast.ILocation;
import org.netkernel.lang.dpml.ast.IReferenceVariable;
import org.netkernel.lang.dpml.ast.IRequestHeader;
import org.netkernel.lang.dpml.ast.IValue;
import org.netkernel.lang.dpml.ast.IValueVariable;
import org.netkernel.lang.dpml.ast.IVariable;
import org.netkernel.lang.dpml.ast.impl.LiteralIndentifier;
import org.netkernel.lang.dpml.ast.impl.LiteralValue;
import org.netkernel.lang.dpml.ast.impl.ValueImpl;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.nkf.impl.NKFRequestImpl;
import org.netkernel.layer0.nkf.impl.NKFResponseReadOnlyImpl;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.util.Utils;
import org.netkernel.request.IExpiry;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.IResponseMeta;
import org.netkernel.request.impl.ExpiryFactory;
import org.netkernel.request.impl.MetaImpl;

import sun.security.action.GetLongAction;

public class DPMLUtils
{
	
	/** 
	 * 
	 * 
	 * @param aArg
	 * @param aContext
	 * @return
	 * @throws NKFException
	 */
	public static IVariable processArgument(IArgument aArg, INKFRequestContext aContext, RuntimeState aState, boolean aEagerDefault) throws NKFException
	{	
		IVariable inVar=aArg.getVariable();
		ILocation inLoc=inVar.getLocation();
		String inName=inVar.getName();
		
		inVar=dereference(inVar,aState,aContext);
		/*
		if (aAlwaysDereference || aArg.getMethod()==IArgument.METHOD_IDENTITY || aArg.getMethod()==IArgument.METHOD_EAGER)
		{	inVar=dereference(inVar,aState,aContext);
		}
		*/
		
		IVariable result=inVar;
		boolean isPrimary=aArg.getName().equals(IArgument.ARG_PRIMARY);
		
		if (!(inVar instanceof IReferenceVariable))
		{	
			if (aArg.getMethod()==IArgument.METHOD_EAGER || isPrimary)
			{	
				if (inVar instanceof IIdentifierVariable)
				{	//source identifier
					String identifier = ((IIdentifierVariable)inVar).getIdentifier(aContext,aState);					
					INKFRequest subReq=aContext.createRequest(identifier);
					INKFResponseReadOnly resp=aContext.issueRequestForResponse(subReq);
					result=new LiteralValue(inName,null,inLoc,resp.getRepresentation(),((NKFResponseReadOnlyImpl)resp).getKernelResponse().getMeta());					
				}
				else // IValueVariable
				{	IValue value = ((IValueVariable)inVar).getValue(aContext,aState,Object.class);
					result=new LiteralValue(inName,null,inLoc,value.getRepresentation(),value.getMeta());							
				}
			}
			else if (aArg.getMethod()==IArgument.METHOD_DEFAULT && aEagerDefault && inVar instanceof IValueVariable)
			{	IValue value = ((IValueVariable)inVar).getValue(aContext,aState,Object.class);
				result=new LiteralValue(inName,null,inLoc,value.getRepresentation(),value.getMeta());	
			}
			else if (aArg.getMethod()==IArgument.METHOD_IDENTITY)
			{	String mime;
				IBinaryStreamRepresentation data;
				if (inVar instanceof IIdentifierVariable)
				{	INKFRequest subReq=aContext.createRequest(((IIdentifierVariable)inVar).getIdentifier(aContext, aState));
					subReq.setRepresentationClass(IBinaryStreamRepresentation.class);
					INKFResponseReadOnly<IBinaryStreamRepresentation> resp=aContext.issueRequestForResponse(subReq);
					mime=resp.getMimeType();
					data=resp.getRepresentation();
				}
				else
				{	IValue value = ((IValueVariable)inVar).getValue(aContext,aState,IBinaryStreamRepresentation.class);
					mime=(String)value.getMeta().getUserMetaData().getValue(IResponseMeta.KEY_MIMETYPE);
					Object rep=value.getRepresentation();
					if (rep instanceof IBinaryStreamRepresentation)
					{	data=(IBinaryStreamRepresentation)rep;
					}
					else
					{	INKFResponseReadOnly<IBinaryStreamRepresentation> respIn=aContext.transreptForResponse(rep, IBinaryStreamRepresentation.class);
						data=respIn.getRepresentation();
						if (mime==null)
						{	mime=respIn.getMimeType();
						}
					}
				}
				
				try
				{	String dataURI=Utils.toDataURI(data, mime);
					result = new LiteralIndentifier(inName,null,inLoc,dataURI,true);
				}
				catch (IOException e)
				{	throw aContext.createFormattedException("MSG_BAD_DATAURI", null, null, e);
				}
				
				//String mime=responseIn.getMimeType();
				//Utils.toDataURI(from, mime);
				/*
				INKFRequest subReq=aContext.createRequest("active:toDataURI");
				if (inVar instanceof IIdentifierVariable)
				{	String identifier = ((IIdentifierVariable)inVar).getIdentifier(aContext,aSubRequestContext);
					subReq.addArgument("operand", identifier);
				}
				else // IValueVariable
				{	IValue value = ((IValueVariable)inVar).getValue(aContext,aSubRequestContext,IBinaryStreamRepresentation.class);
					INKFResponse valResp = createResponseWithMetaData(value,aContext);
					subReq.addArgumentFromResponse("operand", valResp);
				}
				subReq.setRepresentationClass(String.class);
				String rep = (String)aContext.issueRequest(subReq);
				result = new LiteralIndentifier(inName,null,inLoc,rep,true);
				*/
			}
			else
			{	
			}
		}
		return result;
	}	
	
	public static Object processHeader(IRequestHeader aHeader, INKFRequestContext aContext, RuntimeState aState) throws NKFException
	{
		IVariable inVar=aHeader.getVariable();
		Object result;
		if (inVar instanceof IIdentifierVariable)
		{	//source identifier
			String identifier = ((IIdentifierVariable)inVar).getIdentifier(aContext,aState);					
			INKFRequest subReq=aContext.createRequest(identifier);
			INKFResponseReadOnly resp=aContext.issueRequestForResponse(subReq);
			result=resp.getRepresentation();
		}
		else // IValueVariable
		{	IValue value = ((IValueVariable)inVar).getValue(aContext,aState,Object.class);
			result=value.getRepresentation();							
		}
		return result;
	}

	
	public static IVariable dereference(IVariable aVar, RuntimeState aState, INKFRequestContext aContext) throws NKFException
	{
		IVariable result=aVar;
		String name=null;
		IVariable lastDeref=result;
		
		while(result instanceof IReferenceVariable)
		{	lastDeref=result;
			result=((IReferenceVariable)result).resolve(aContext,aState);
		}
		if (result==null)
		{	throw aContext.createFormattedException("EX_DPML_RUNTIME","MSG_RUN_NOT_FOUND",formatLocationMessage(lastDeref,aContext),null,name);
		}
		return result;
	}	
	
	public static String formatLocationMessage(IVariable aVar, INKFRequestContext aContext)
	{	ILocation loc = aVar.getLocation();
		if (loc!=null)
		{	return aContext.formatMessage("MSG_PXML_LOCATION",new Object[] {loc.getSourceIdentity(),new Integer(loc.getLine()),new Integer(loc.getColumn())});
		}
		else
		{	return null;
		}
	}
	
	public static INKFResponse createResponseWithMetaData(IValue aValue, INKFRequestContext aContext)
	{	INKFResponse result= aContext.createResponseFrom(aValue.getRepresentation());
		IResponseMeta meta=aValue.getMeta();
		IRequestResponseFields headers = meta.getUserMetaData();
		if (headers.getValue(ResponseAccessor.HEADER_EXPIRY)!=null || headers.getValue(IResponseMeta.KEY_NO_CACHE)!=null)
		{	//need to remove this header to stop it propagating up and effecting someone else
			int s=headers.size();
			for (int i=0; i<s; i++)
			{	String key=headers.getKey(i);
				if (!key.equals(ResponseAccessor.HEADER_EXPIRY) && !key.equals(IResponseMeta.KEY_NO_CACHE))
				{	result.setHeader(key, headers.getValue(i));
				}
			}
		}
		else
		{	result.setHeaders(headers);
		}
		return result;
	}	
	
	public static void setResponseExpiry(INKFResponse aResponse, IRequestResponseFields aResponseHeaders, INKFResponseReadOnly aScriptResponse)
	{
		Integer modifiedExpiry=(Integer)aResponseHeaders.getValue(ResponseAccessor.HEADER_EXPIRY);
		Long modifiedExpiryPeriod=(Long)aResponseHeaders.getValue(ResponseAccessor.HEADER_EXPIRY_PERIOD);
		final IExpiry scriptExpiry=((NKFResponseReadOnlyImpl)aScriptResponse).getMeta().getExpiry();
		
		switch(modifiedExpiry)
		{	case INKFResponse.EXPIRY_ALWAYS:
				aResponse.setExpiry(INKFResponse.EXPIRY_ALWAYS);
				break;
			case INKFResponse.EXPIRY_NEVER:
				aResponse.setExpiry(INKFResponse.EXPIRY_NEVER);
				break;
			case INKFResponse.EXPIRY_DEPENDENT:
				aResponse.setExpiry(INKFResponse.EXPIRY_DEPENDENT);
				break;
			case INKFResponse.EXPIRY_CONSTANT:
				aResponse.setExpiry(INKFResponse.EXPIRY_CONSTANT,modifiedExpiryPeriod+System.currentTimeMillis());
				break;
			case INKFResponse.EXPIRY_MAX_CONSTANT_DEPENDENT:
				aResponse.setExpiry(INKFResponse.EXPIRY_MAX_CONSTANT_DEPENDENT,modifiedExpiryPeriod+System.currentTimeMillis());
				break;
			case INKFResponse.EXPIRY_MIN_CONSTANT_DEPENDENT:
				aResponse.setExpiry(INKFResponse.EXPIRY_MIN_CONSTANT_DEPENDENT,modifiedExpiryPeriod+System.currentTimeMillis());
				break;
			default:
				//invalid value set
		}
	}
	
	public static IValue requestIdentifier(IIdentifierVariable aVar, INKFRequestContext aContext, RuntimeState aState, Class aRepresentationClass) throws NKFException
	{
		String id=aVar.getIdentifier(aContext, aState);
		INKFRequest req=aContext.createRequest(id);
		req.setVerb(aContext.getThisRequest().getVerb());
		//req.setRequestScope(aSubRequestContext);
		req.setRepresentationClass(aRepresentationClass);
		INKFResponseReadOnly r=aContext.issueRequestForResponse(req);
		return new ValueImpl(r.getRepresentation(),((NKFResponseReadOnlyImpl)r).getMeta());		
	}
	
	public static IVariable createArgumentLiteralValue(String aName, Object aRepresentation, IResponseMeta aMeta, ILocation aLocation)
	{	IResponseMeta meta=aMeta!=null?aMeta:new MetaImpl(ExpiryFactory.getDependentExpiry());
		return new LiteralValue("arg:"+aName, null, aLocation, aRepresentation, meta);
	}	
	
	public static Object ensureClassOfValue(IValue aValue, Class aRepresentationClass, INKFRequestContext aContext) throws NKFException
	{
		Object result;
		if (aRepresentationClass.isAssignableFrom(aValue.getRepresentation().getClass()))
		{	result=aValue.getRepresentation();
		}
		else
		{	INKFResponse valResp = createResponseWithMetaData(aValue,aContext);
			String id=SimpleIdentifierImpl.getUnique("literal:var").toString();
			INKFRequest subReq=aContext.createRequest(id);
			subReq.setVerb(INKFRequestReadOnly.VERB_TRANSREPRESENT);
			subReq.addPrimaryArgumentFromResponse(valResp);
			subReq.setRepresentationClass(aRepresentationClass);
			result=aContext.issueRequest(subReq);
		}
		return result;
	}	

}
